@@ -24,14 +24,14 @@ balancesheetSpec = CompoundBalanceCommandSpec {
   cbcqueries  = [
      CBCSubreportSpec{
       cbcsubreporttitle="Assets"
-     ,cbcsubreportquery=journalAssetAccountQuery
+     ,cbcsubreportquery=Type [Asset]
      ,cbcsubreportoptions=(\ropts -> ropts{normalbalance_=Just NormallyPositive})
      ,cbcsubreporttransform=id
      ,cbcsubreportincreasestotal=True
      }
     ,CBCSubreportSpec{
       cbcsubreporttitle="Liabilities"
-     ,cbcsubreportquery=journalLiabilityAccountQuery
+     ,cbcsubreportquery=Type [Liability]
      ,cbcsubreportoptions=(\ropts -> ropts{normalbalance_=Just NormallyNegative})
      ,cbcsubreporttransform=fmap maNegate
      ,cbcsubreportincreasestotal=False
