@@ -38,9 +38,6 @@ module Hledger.Cli.Commands.Demo (
  ,demo
 ) where
 
-import Hledger
-import Hledger.Cli.CliOptions
-import System.Exit (exitFailure)
 import Text.Printf
 import Control.Concurrent (threadDelay)
 import System.Process (callProcess)
@@ -50,11 +47,17 @@ import Data.List (isPrefixOf, find, findIndex, isInfixOf, dropWhileEnd)
 import Control.Applicative ((<|>))
 import Data.ByteString as B (ByteString)
 import Data.Maybe
-import qualified Data.ByteString.Char8 as B
+import Data.ByteString.Char8 qualified as B
+import Safe (tailMay)
 import System.IO.Temp (withSystemTempFile)
 import System.IO (hClose)
 import System.Console.CmdArgs.Explicit (flagReq)
 
+import Hledger
+import Hledger.Cli.CliOptions
+import System.Directory (findExecutable)
+import Control.Monad (when)
+
 demos :: [Demo]
 demos = map readDemo [
   -- XXX these are confusing, redo
@@ -84,7 +87,8 @@ demomode = hledgerCommandMode
   []
   ([], Just $ argsFlag optsstr)
 
-optsstr = "[NUM|PREFIX|SUBSTR] [-- ASCIINEMAOPTS]"
+optsstr = "[NUM|PREFIX|SUBSTR]"
+-- optsstr = "[NUM|PREFIX|SUBSTR] [-- ASCIINEMAOPTS]"
 usagestr = "Usage: hledger demo " <> optsstr
 
 -- | The demo command.
@@ -92,15 +96,18 @@ demo :: CliOpts -> Journal -> IO ()
 demo CliOpts{rawopts_=rawopts, reportspec_=ReportSpec{_rsQuery=_query}} _j = do
   -- demos <- getCurrentDirectory >>= readDemos
   case listofstringopt "args" rawopts of
-    [] -> putStrLn usagestr >> printDemos
+    [] -> putStrLn usagestr >> putStr listDemos
     (a:as) ->
       case findDemo demos a of
-        Nothing -> do
-          putStrLn $ "No demo \"" <> a <> "\" was found."
-          putStrLn usagestr
-          printDemos
-          exitFailure
+        Nothing -> error' $ unlines
+          ["No demo \"" <> a <> "\" was found."
+          ,usagestr
+          ,listDemos
+          ]
         Just (Demo t c) -> do
+          -- check if asciinema is installed, first
+          masciinema <- findExecutable "asciinema"
+          when (isNothing masciinema) $ error' "Could not find 'asciinema'; please install that first."
           let
             -- try to preserve the original pauses a bit while also moving things along
             defidlelimit = 10
@@ -117,6 +124,7 @@ demo CliOpts{rawopts_=rawopts, reportspec_=ReportSpec{_rsQuery=_query}} _j = do
           putStrLn line
           putStrLn ""
           threadDelay 1000000
+          -- XXX this used to see asciinema options after --, currently it doesn't
           runAsciinemaPlay speed defidlelimit c as
           putStrLn ""
           putStrLn line
@@ -128,8 +136,7 @@ readDemo content = Demo title content
       where
         readTitle s
           | "\"title\":" `isPrefixOf` s = takeWhile (/='"') $ drop 1 $ lstrip $ drop 8 s
-          | null s = ""
-          | otherwise = readTitle $ tail s
+          | otherwise = maybe "" readTitle $ tailMay s
 
 findDemo :: [Demo] -> String -> Maybe Demo
 findDemo ds s =
@@ -139,16 +146,16 @@ findDemo ds s =
   where
     sl = lowercase s
 
-printDemos :: IO ()
-printDemos = putStrLn $ unlines $
+listDemos :: String
+listDemos = unlines $
   "Demos:" :
   -- "" :
   [show i <> ") " <> bold' t | (i, Demo t _) <- zip [(1::Int)..] demos]
 
 -- | Run asciinema play with the given speed and idle limit, passing the given content to its stdin.
 runAsciinemaPlay :: Float -> Float -> ByteString -> [String] -> IO ()
-runAsciinemaPlay speed idlelimit content args =
-    -- XXX try piping to stdin also
+runAsciinemaPlay speed idlelimit content args = do
+  -- XXX try piping to stdin also
   withSystemTempFile "hledger-cast" $ \f h -> do
     -- don't add an extra newline here, it breaks asciinema 2.3.0 (#2094).
     -- XXX we could try harder and strip excess newlines/carriage returns+linefeeds here
@@ -160,12 +167,14 @@ runAsciinemaPlay speed idlelimit content args =
       ,[f]
       ,args
       ])
-      `catchIOError` \err -> do
-        putStrLn $ "\n" <> show err
-        putStrLn "Error: running asciinema failed. Trying 'asciinema --version':"
-        callProcess "asciinema" ["--version"] `catchIOError` \_ ->
-          putStrLn "This also failed. Check that asciinema is installed in your PATH."
-        exitFailure
+    `catchIOError` \err -> do
+      printError $ unlines
+        [""
+        ,show err
+        ,"Running asciinema failed. Trying 'asciinema --version':"
+        ]
+      callProcess "asciinema" ["--version"]
+      `catchIOError` \_ -> error' "This also failed."
   where
     showwithouttrailingzero = dropWhileEnd (=='.') . dropWhileEnd (=='0') . show
 
