@@ -20,28 +20,35 @@ module Hledger.Cli.Commands.Register (
 
 import Data.Default (def)
 import Data.Maybe (fromMaybe, isJust)
-import qualified Data.Text as T
-import qualified Data.Text.Lazy as TL
-import qualified Data.Text.Lazy.IO as TL
-import qualified Data.Text.Lazy.Builder as TB
+import Data.Text (Text)
+import Data.Map qualified as Map
+import Data.Text qualified as T
+import Data.Text.Lazy qualified as TL
+import Data.Text.Lazy.IO qualified as TL
+import Data.Text.Lazy.Builder qualified as TB
 import System.Console.CmdArgs.Explicit (flagNone, flagReq)
 
 import Hledger hiding (per)
-import Hledger.Read.CsvUtils (CSV, CsvRecord, printCSV, printTSV)
+import Hledger.Write.Csv (CSV, printCSV, printTSV)
+import Hledger.Write.Ods (printFods)
+import Hledger.Write.Html.Lucid (styledTableHtml)
+import Hledger.Write.Spreadsheet qualified as Spr
 import Hledger.Cli.CliOptions
 import Hledger.Cli.Utils
-import Text.Tabular.AsciiWide hiding (render)
+import Hledger.Cli.Anchor (setAccountAnchor, dateCell)
+import Text.Tabular.AsciiWide (Cell(..), Align(..), Properties(..), Header(Header, Group), renderRowB, textCell, tableBorders, borderSpaces)
+import Lucid qualified
 import Data.List (sortBy)
 import Data.Char (toUpper)
 import Data.List.Extra (intersect)
-import System.Exit (exitFailure)
+import System.IO qualified as IO
 
 registermode = hledgerCommandMode
   $(embedFileRelative "Hledger/Cli/Commands/Register.txt")
   ([flagNone ["cumulative"] (setboolopt "cumulative")
-     "show running total from report start date (default)"
+     (accumprefix ++ "show running total from report start date (default)")
   ,flagNone ["historical","H"] (setboolopt "historical")
-     "show historical running total/balance (includes postings before report start date)\n "
+     (accumprefix ++ "show historical running total/balance (includes postings before report start date)")
   ,flagNone ["average","A"] (setboolopt "average")
      "show running average of posting amounts instead of total (implies --empty)"
   ,let arg = "DESC" in
@@ -49,6 +56,9 @@ registermode = hledgerCommandMode
     ("fuzzy search for one recent posting with description closest to "++arg)
   ,flagNone ["related","r"] (setboolopt "related") "show postings' siblings instead"
   ,flagNone ["invert"] (setboolopt "invert") "display all amounts with reversed sign"
+  ,flagReq  ["sort"] (\s opts -> Right $ setopt "sort" s opts) "FIELDS" 
+    ("sort by: " <> sortKeysDescription
+    <> ", or a comma-separated combination of these. For a descending sort, prefix with -. (Default: date)")
   ,flagReq  ["width","w"] (\s opts -> Right $ setopt "width" s opts) "N"
      ("set output width (default: " ++
 #ifdef mingw32_HOST_OS
@@ -56,15 +66,18 @@ registermode = hledgerCommandMode
 #else
       "terminal width"
 #endif
-      ++ " or $COLUMNS). -wN,M sets description width as well."
+      ++ "). -wN,M sets description width as well."
      )
   ,flagNone ["align-all"] (setboolopt "align-all") "guarantee alignment across all lines (slower)"
-  ,outputFormatFlag ["txt","csv","tsv","json"]
+  ,flagReq  ["base-url"] (\s opts -> Right $ setopt "base-url" s opts) "URLPREFIX" "in html output, generate links to hledger-web, with this prefix. (Usually the base url shown by hledger-web; can also be relative.)"
+  ,outputFormatFlag ["txt","csv","tsv","html","fods","json"]
   ,outputFileFlag
   ])
-  [generalflagsgroup1]
+  cligeneralflagsgroups1
   hiddenflags
   ([], Just $ argsFlag "[QUERY]")
+  where
+    accumprefix = "accumulation mode: "
 
 -- | Print a (posting) register report.
 register :: CliOpts -> Journal -> IO ()
@@ -74,7 +87,7 @@ register opts@CliOpts{rawopts_=rawopts, reportspec_=rspec} j
   | Just desc <- maybestringopt "match" rawopts = do
       let ps = [p | (_,_,_,p,_) <- rpt]
       case similarPosting ps desc of
-        Nothing -> putStrLn "no matches found." >> exitFailure
+        Nothing -> error' $ "no postings found with description like " <> show desc
         Just p  -> TL.putStr $ postingsReportAsText opts [pri]
                   where pri = (Just (postingDate p)
                               ,Nothing
@@ -87,23 +100,57 @@ register opts@CliOpts{rawopts_=rawopts, reportspec_=rspec} j
     styles = journalCommodityStylesWith HardRounding j
     rpt = postingsReport rspec j
     render | fmt=="txt"  = postingsReportAsText opts
+           | fmt=="json" = toJsonText
            | fmt=="csv"  = printCSV . postingsReportAsCsv
            | fmt=="tsv"  = printTSV . postingsReportAsCsv
-           | fmt=="json" = toJsonText
+           | fmt=="html" =
+                (<>"\n") . Lucid.renderText . styledTableHtml .
+                map (map (fmap Lucid.toHtml)) .
+                postingsReportAsSpreadsheet oneLineNoCostFmt baseUrl query
+           | fmt=="fods" =
+                printFods IO.localeEncoding . Map.singleton "Register" .
+                (,) (1,0) .
+                postingsReportAsSpreadsheet oneLineNoCostFmt baseUrl query
            | otherwise   = error' $ unsupportedOutputFormatError fmt  -- PARTIAL:
       where fmt = outputFormatFromOpts opts
+            baseUrl = balance_base_url_ $ _rsReportOpts rspec
+            query = querystring_ $ _rsReportOpts rspec
 
 postingsReportAsCsv :: PostingsReport -> CSV
-postingsReportAsCsv is =
-  ["txnidx","date","code","description","account","amount","total"]
-  :
-  map postingsReportItemAsCsvRecord is
+postingsReportAsCsv =
+  Spr.rawTableContent . postingsReportAsSpreadsheet machineFmt Nothing []
 
-postingsReportItemAsCsvRecord :: PostingsReportItem -> CsvRecord
-postingsReportItemAsCsvRecord (_, _, _, p, b) = [idx,date,code,desc,acct,amt,bal]
+-- ToDo: --layout=bare etc.
+-- ToDo: Text output does not show headers, but Spreadsheet does
+postingsReportAsSpreadsheet ::
+  AmountFormat -> Maybe Text -> [Text] ->
+  PostingsReport -> [[Spr.Cell Spr.NumLines Text]]
+postingsReportAsSpreadsheet fmt baseUrl query is =
+  Spr.addHeaderBorders
+    (map Spr.headerCell
+      ["txnidx","date","code","description","account","amount","total"])
+  :
+  map (postingsReportItemAsRecord fmt baseUrl query) is
+
+{- ToDo:
+link txnidx to journal URL,
+   however, requires Web.Widget.Common.transactionFragment
+-}
+postingsReportItemAsRecord ::
+    (Spr.Lines border) =>
+    AmountFormat -> Maybe Text -> [Text] ->
+    PostingsReportItem -> [Spr.Cell border Text]
+postingsReportItemAsRecord fmt baseUrl query (_, _, _, p, b) =
+    [idx,
+     (dateCell baseUrl query (paccount p) date) {Spr.cellType = Spr.TypeDate},
+     cell code, cell desc,
+     setAccountAnchor baseUrl query (paccount p) $ cell acct,
+     amountCell (pamount p),
+     amountCell b]
   where
-    idx  = T.pack . show . maybe 0 tindex $ ptransaction p
-    date = showDate $ postingDate p -- XXX csv should show date2 with --date2
+    cell = Spr.defaultCell
+    idx  = Spr.integerCell . maybe 0 tindex $ ptransaction p
+    date = postingDate p -- XXX csv should show date2 with --date2
     code = maybe "" tcode $ ptransaction p
     desc = maybe "" tdescription $ ptransaction p
     acct = bracket $ paccount p
@@ -113,8 +160,8 @@ postingsReportItemAsCsvRecord (_, _, _, p, b) = [idx,date,code,desc,acct,amt,bal
                              VirtualPosting -> wrap "(" ")"
                              _ -> id
     -- Since postingsReport strips prices from all Amounts when not used, we can display prices.
-    amt = wbToText . showMixedAmountB csvDisplay $ pamount p
-    bal = wbToText $ showMixedAmountB csvDisplay b
+    amountCell amt =
+      wbToText <$> Spr.cellFromMixedAmount fmt (Spr.Class "amount", amt)
 
 -- | Render a register report as plain text suitable for console output.
 postingsReportAsText :: CliOpts -> PostingsReport -> TL.Text
@@ -258,7 +305,7 @@ tests_Register = testGroup "Register" [
 
    testGroup "postingsReportAsText" [
     testCase "unicode in register layout" $ do
-      j <- readJournal' "2009/01/01 * медвежья шкура\n  расходы:покупки  100\n  актив:наличные\n"
+      j <- readJournal'' "2009/01/01 * медвежья шкура\n  расходы:покупки  100\n  актив:наличные\n"
       let rspec = defreportspec
       (TL.unpack . postingsReportAsText defcliopts $ postingsReport rspec j)
         @?=
