@@ -16,17 +16,16 @@ module Hledger.Cli.Commands.Roi (
 ) where
 
 import Control.Monad
-import System.Exit
 import Data.Time.Calendar
 import Text.Printf
 import Data.Bifunctor (second)
-import Data.Either (fromLeft, fromRight, isLeft)
 import Data.Function (on)
 import Data.List
 import Numeric.RootFinding
 import Data.Decimal
-import qualified Data.Text as T
-import qualified Data.Text.Lazy.IO as TL
+import Data.Text qualified as T
+import Data.Text.Lazy.IO qualified as TL
+import Safe (headDef)
 import System.Console.CmdArgs.Explicit as CmdArgs
 
 import Text.Tabular.AsciiWide as Tab
@@ -43,7 +42,7 @@ roimode = hledgerCommandMode
   ,flagReq ["profit-loss","pnl"] (\s opts -> Right $ setopt "pnl" s opts) "QUERY"
     "query to select profit-and-loss or appreciation/valuation transactions"
   ]
-  [generalflagsgroup1]
+  cligeneralflagsgroups1
   hiddenflags
   ([], Just $ argsFlag "[QUERY]")
 
@@ -92,36 +91,35 @@ roi CliOpts{rawopts_=rawopts, reportspec_=rspec@ReportSpec{_rsReportOpts=ReportO
     filteredj = filterJournalTransactions investmentsQuery j
     trans = dbg3 "investments" $ jtxns filteredj
 
-  when (null trans) $ do
-    putStrLn "No relevant transactions found. Check your investments query"
-    exitFailure
+  when (null trans) $
+    error' "No relevant transactions found. Check your investments query"
 
-  let (fullPeriod, spans) = reportSpan filteredj rspec
+  let (fullPeriodDateSpan, mspans) = reportSpan filteredj rspec
 
-  let priceDirectiveDates = dbg3 "priceDirectiveDates" $ map pddate $ jpricedirectives j
+  let err = error' "Undefined start or end of the period - will be unable to compute the rates of return"
+      spans = maybe err (map (second (addDays 1)) . dayPartitionToList) mspans
+      fullPeriod = case fullPeriodDateSpan of
+        DateSpan (Just b) (Just e) -> (fromEFDay b, fromEFDay e)
+        _ -> err
 
-  let processSpan (DateSpan Nothing _) = error "Undefined start of the period - will be unable to compute the rates of return"
-      processSpan (DateSpan _ Nothing) = error "Undefined end of the period - will be unable to compute the rates of return"
-      processSpan spn@(DateSpan (Just begin) (Just end)) = do
+  let processSpan (b, e) = do
         -- Spans are [begin,end), and end is 1 day after the actual end date we are interested in
         let
-          b = fromEFDay begin
-          e = fromEFDay end
+          spn = DateSpan (Just $ Exact b) (Just $ Exact e)
+
           cashFlowApplyCostValue = map (\(d,amt) -> (d,mixedAmountValue e d amt))
 
-          valueBefore =
+          valueBefore = dbg3 "valueBefore" $
             mixedAmountValue e b $
             total trans (And [ investmentsQuery
-                             , Date (DateSpan Nothing (Just begin))])
+                             , Date (DateSpan Nothing (Just $ Exact b))])
 
-          valueAfter  =
+          valueAfter  = dbg3 "valueAfter" $
             mixedAmountValue e e $
             total trans (And [investmentsQuery
-                             , Date (DateSpan Nothing (Just end))])
+                             , Date (DateSpan Nothing (Just $ Exact e))])
 
-          priceDates = dbg3 "priceDates" $ nub $ filter (spanContainsDate spn) priceDirectiveDates
-          cashFlow =
-            ((map (,nullmixedamt) priceDates)++) $
+          cashFlow = dbg3 "cashFlow" $
             cashFlowApplyCostValue $
             calculateCashFlow wd trans (And [ Not investmentsQuery
                                             , Not pnlQuery
@@ -142,23 +140,25 @@ roi CliOpts{rawopts_=rawopts, reportspec_=rspec@ReportSpec{_rsReportOpts=ReportO
         let smallIsZero x = if abs x < 0.01 then 0.0 else x
         return [ showDate b
                , showDate (addDays (-1) e)
-               , T.pack $ showMixedAmount $ styleAmounts styles $ valueBefore
-               , T.pack $ showMixedAmount $ styleAmounts styles $ cashFlowAmt
+               , T.pack $ showMixedAmountOneLineWithoutCost False $ styleAmounts styles $ valueBefore
+               , T.pack $ showMixedAmountOneLineWithoutCost False $ styleAmounts styles $ cashFlowAmt
                -- , T.pack $ showMixedAmount $
                --   -- dbg0With (lbl "cashflow after styling".showMixedAmountOneLine) $
-               --   mapMixedAmount (amountSetFullPrecisionOr (Just defaultMaxPrecision)) $
+               --   mapMixedAmount (amountSetFullPrecisionUpTo (Just defaultMaxPrecision)) $
                --   styleAmounts (styles
                --                 -- & dbg0With (lbl "styles".show))
                --   cashFlowAmt
                --   -- & dbg0With (lbl "cashflow before styling".showMixedAmountOneLine)
-               , T.pack $ showMixedAmount $ styleAmounts styles $ valueAfter
-               , T.pack $ showMixedAmount $ styleAmounts styles $ (valueAfter `maMinus` (valueBefore `maPlus` cashFlowAmt))
+               , T.pack $ showMixedAmountOneLineWithoutCost False $ styleAmounts styles $ valueAfter
+               , T.pack $ showMixedAmountOneLineWithoutCost False $ styleAmounts styles $ (valueAfter `maMinus` (valueBefore `maPlus` cashFlowAmt))
                , T.pack $ printf "%0.2f%%" $ smallIsZero irr
                , T.pack $ printf "%0.2f%%" $ smallIsZero periodTwr
                , T.pack $ printf "%0.2f%%" $ smallIsZero annualizedTwr ]
 
   periodRows <- forM spans processSpan
-  totalRow <- processSpan fullPeriod
+  totalRow <- case periodRows of
+    [singleRow] -> return singleRow
+    _           -> processSpan fullPeriod
 
   let rowTitles = Tab.Group Tab.NoLine (map (Header . T.pack . show) (take (length periodRows) [1..]))
 
@@ -178,114 +178,99 @@ roi CliOpts{rawopts_=rawopts, reportspec_=rspec@ReportSpec{_rsReportOpts=ReportO
 
   TL.putStrLn $ Tab.render prettyTables id id id table
 
-timeWeightedReturn styles showCashFlow prettyTables investmentsQuery trans mixedAmountValue (OneSpan begin end valueBeforeAmt valueAfter cashFlow pnl) = do
-  let valueBefore = unMix valueBeforeAmt
-  let initialUnitPrice = 100 :: Decimal
-  let initialUnits = valueBefore / initialUnitPrice
-  let changes =
-        -- If cash flow and PnL changes happen on the same day, this
-        -- will sort PnL changes to come before cash flows (on any
-        -- given day), so that we will have better unit price computed
-        -- first for processing cash flow. This is why pnl changes are Left
-        -- and cashflows are Right.
-        -- However, if the very first date in the changes list has both
-        -- PnL and CashFlow, we would not be able to apply pnl change to 0 unit,
-        -- which would lead to an error. We make sure that we have at least one
-        -- cashflow entry at the front, and we know that there would be at most
-        -- one for the given date, by construction. Empty CashFlows added
-        -- because of a begin date before the first transaction are not seen as
-        -- a valid cashflow entry at the front.
-        zeroUnitsNeedsCashflowAtTheFront
+-- Entry for TWR computation, capturing all cashflows that are potentially accompanied by pnl change on the same day (if not, it is zero)
+data TwrPeriod = TwrPeriod { twrStartDate :: Day, twrEndDate :: Day, twrStartValue :: Decimal, twrValueBeforeCashflow :: Decimal, twrPnl :: Decimal, twrCashflow :: Decimal, twrValueAfterCashflow :: Decimal } deriving (Eq, Show)
+
+timeWeightedReturn _styles showCashFlow prettyTables investmentsQuery trans mixedAmountValue (OneSpan begin end valueBeforeAmt valueAfterAmt cashflows pnls) = do
+  let datedCashflows =
+        -- Aggregate all entries for a single day, assuming that intraday interest is negligible
+        dbg3 "datedCashflows"
         $ sort
-        $ datedCashflows ++ datedPnls
-        where
-          zeroUnitsNeedsCashflowAtTheFront changes1 =
-            if initialUnits > 0 then changes1
-            else
-              let (leadingEmptyCashFlows, rest) = span isEmptyCashflow changes1
-                  (leadingPnls, rest') = span (isLeft . snd) rest
-                  (firstCashflow, rest'') = splitAt 1 rest'
-              in leadingEmptyCashFlows ++ firstCashflow ++ leadingPnls ++ rest''
+        $ map (\datecashes -> let (dates, cash) = unzip datecashes in (headDef (error' "Roi.hs: datecashes was null, please report a bug") dates, maSum cash))
+        $ groupBy ((==) `on` fst)
+        $ sortOn fst
+        $ map (second maNegate)
+        $ cashflows
 
-          isEmptyCashflow (_date, amt) = case amt of
-            Right amt' -> mixedAmountIsZero amt'
-            Left _     -> False
+      valueBefore = dbg3 ("value at the start of the interval, "++show begin) $ unMix valueBeforeAmt
+      valueAfter = dbg3 ("value at the end of the interval, "++show end) $ unMix valueAfterAmt
+      
+      investmentPostings = concatMap (filter (matchesPosting investmentsQuery) . realPostings) trans
 
-          datedPnls = map (second Left) $ aggregateByDate pnl
+      totalInvestmentPostingsTill date = sumPostings $ filter (matchesPosting (Date (DateSpan Nothing (Just $ Exact date)))) investmentPostings
 
-          datedCashflows = map (second Right) $ aggregateByDate cashFlow
+      -- filter span is (-infinity, date+1), which gives us effectively (-infinity, date]
+      valueAfterDate date = unMix $ mixedAmountValue end date $ totalInvestmentPostingsTill (addDays 1 date)
 
-          aggregateByDate datedAmounts =
-            -- Aggregate all entries for a single day, assuming that intraday interest is negligible
-            sort
-            $ map (\date_cash -> let (dates, cash) = unzip date_cash in (head dates, maSum cash))
-            $ groupBy ((==) `on` fst)
-            $ sortOn fst
-            $ map (second maNegate)
-            $ datedAmounts
+      pnlOn date = unMix $ maNegate $ sum $ map snd $ filter ((==date).fst) pnls
 
-  let units =
-        tail $
-        scanl
-          (\(_, _, unitPrice, unitBalance) (date, amt) ->
-             let valueOnDate = unMix $ mixedAmountValue end date $ total trans (And [investmentsQuery, Date (DateSpan Nothing (Just $ Exact date))])
-             in
-             case amt of
-               Right amt' ->
-                 -- we are buying or selling
-                 let unitsBoughtOrSold = unMix amt' / unitPrice
-                 in (valueOnDate, unitsBoughtOrSold, unitPrice, unitBalance + unitsBoughtOrSold)
-               Left pnl' ->
-                 -- PnL change
-                 let valueAfterDate = valueOnDate + unMix pnl'
-                     unitPrice' = valueAfterDate/unitBalance
-                 in (valueOnDate, 0, unitPrice', unitBalance))
-          (0, 0, initialUnitPrice, initialUnits)
-          $ dbg3 "changes" changes
+  -- We are dividing the period [begin, end) into subperiods on each cashflow, and then compute
+  -- the rate of return for each subperiod. For this we need to know the value of the investment
+  -- at the beginning and end of each subperiod, adjusted for cashflow.
+  -- 
+  -- Subperiods are going to be [valueBefore ... (c_0,v_0)][... (c_1, v_1)][... (c_2,v_2)] ... [... (c_n,v_n)][... valueAfter]
+  -- , where v_i is the value of investment computed immediately after cashflow c_i
+  --
+  -- Calculate interest for each subperiod, adjusting the value at the start of the period by the cashflow
+  -- For subperiods [valueBefore ... (c_0,v_0)][... (c_1, v_1)][... (c_2,v_2)] ... [... (c_n,v_n)][... valueAfter], the computation is going to be
+  -- 1 + twr = (v_0 - c_0)/valueBefore + (v_1 - c_1) / v_0 +  ... + valueAfter/v_n
+  -- See https://en.wikipedia.org/wiki/Time-weighted_return#Time-weighted_return_compensating_for_external_flows
+  let calculateSubPeriods (startDate,startValue) [] =
+        let subPeriodReturn =
+              if startValue == 0 || valueAfter == 0
+              then 0
+              else valueAfter/startValue - 1
+        in
+        [(subPeriodReturn, TwrPeriod startDate end startValue valueAfter 0 0 valueAfter)]
+      calculateSubPeriods (startDate,startValue) ((date,cashflow):rest) =
+        let (valueBeforeCashflow, valueAfterCashflow, pnl) =
+              let valueAfterPrevDay = valueAfterDate (addDays (-1) date)
+                  pnlOnDay = pnlOn date
+              in
+                -- If value was zero at the start of the period, then any PnL on cashflow date would accrue after it, not before.
+                -- If there was some value already, we can assume that PnL contributes to this period's rate
+                if startValue == 0
+                then (valueAfterPrevDay, valueAfterDate date - pnlOnDay, 0)
+                else (valueAfterPrevDay + pnl, valueAfterDate date, pnlOnDay)
+            subPeriodReturn =
+              if valueBeforeCashflow == 0 || startValue == 0
+              then 0
+              else (valueBeforeCashflow / startValue) - 1
+        in
+        (subPeriodReturn, (TwrPeriod startDate date startValue valueBeforeCashflow pnl (unMix cashflow) valueAfterCashflow)) : calculateSubPeriods (date,valueAfterCashflow) rest
 
-  let finalUnitBalance = if null units then initialUnits else let (_,_,_,u) = last units in u
-      finalUnitPrice = if finalUnitBalance == 0 then
-                         if null units then initialUnitPrice
-                         else let (_,_,lastUnitPrice,_) = last units in lastUnitPrice
-                       else (unMix valueAfter) / finalUnitBalance
-      -- Technically, totalTWR should be (100*(finalUnitPrice - initialUnitPrice) / initialUnitPrice), but initalUnitPrice is 100, so 100/100 == 1
-      totalTWR = roundTo 2 $ (finalUnitPrice - initialUnitPrice)
+  let subPeriods = dbg3 "subPeriods" $ calculateSubPeriods (begin,valueBefore) datedCashflows
+
+  -- Compute overall time-weighted rate of return
+  let twr =
+        dbg3 "twr" $
+        if subPeriods == []
+        then if valueBefore == 0 then 0 else (valueAfter - valueBefore)/valueBefore
+        else foldl (\acc periodRate -> (1+acc)*(1+periodRate)-1) 0 (map fst subPeriods)
       (startYear, _, _) = toGregorian begin
       years = fromIntegral (diffDays end begin) / (if isLeapYear startYear then 366 else 365) :: Double
-      annualizedTWR = 100*((1+(realToFrac totalTWR/100))**(1/years)-1) :: Double
+      periodTWR = roundTo 2 $ 100 * twr
+      annualizedTWR = 100*((1+(realToFrac twr))**(1/years)-1) :: Double
 
   when showCashFlow $ do
-    printf "\nTWR cash flow for %s - %s\n" (showDate begin) (showDate (addDays (-1) end))
-    let (dates', amts) = unzip changes
-        cashflows' = map (fromRight nullmixedamt) amts
-        pnls = map (fromLeft nullmixedamt) amts
-        (valuesOnDate,unitsBoughtOrSold', unitPrices', unitBalances') = unzip4 units
-        add x lst = if valueBefore/=0 then x:lst else lst
-        dates = add begin dates'
-        cashflows = add valueBeforeAmt cashflows'
-        unitsBoughtOrSold = add initialUnits unitsBoughtOrSold'
-        unitPrices = add initialUnitPrice unitPrices'
-        unitBalances = add initialUnits unitBalances'
-
-    TL.putStr $ Tab.render prettyTables id id T.pack
+    printf "\nTWR cash flow entries and subperiod rates for period %s - %s\n" (showDate begin) (showDate (addDays (-1) end))
+    let showDecimalT = T.pack . showDecimal
+    TL.putStr $ Tab.render prettyTables T.pack id id
       (Table
-       (Tab.Group NoLine (map (Header . showDate) dates))
-       (Tab.Group DoubleLine [ Tab.Group Tab.SingleLine [Tab.Header "Portfolio value", Tab.Header "Unit balance"]
-                         , Tab.Group Tab.SingleLine [Tab.Header "Pnl", Tab.Header "Cashflow", Tab.Header "Unit price", Tab.Header "Units"]
-                         , Tab.Group Tab.SingleLine [Tab.Header "New Unit Balance"]])
-       [ [val, oldBalance, pnl', cashflow, prc, udelta, balance]
-       | val <- map showDecimal valuesOnDate
-       | oldBalance <- map showDecimal (0:unitBalances)
-       | balance <- map showDecimal unitBalances
-       | pnl' <- map (showMixedAmount . styleAmounts styles) pnls
-       | cashflow <- map (showMixedAmount . styleAmounts styles) cashflows
-       | prc <- map showDecimal unitPrices
-       | udelta <- map showDecimal unitsBoughtOrSold ])
+       (Tab.Group Tab.NoLine [ Header (show n) | n <-[1..length subPeriods]])
+       (Tab.Group DoubleLine [ Tab.Group Tab.SingleLine [Tab.Header "Subperiod start", Tab.Header "Cashflow date"]
+                             , Tab.Group Tab.SingleLine [Tab.Header "Value at start", Tab.Header "Value before cashflow (inc PnL)", Tab.Header "PnL on day", Tab.Header "Cashflow",  Tab.Header "Value after cashflow"]
+                             , Tab.Group Tab.SingleLine [Tab.Header "Subperiod rate, %"]])
+       [ [ showDate (twrStartDate sp), showDate (twrEndDate sp)
+         , showDecimalT (twrStartValue sp), showDecimalT (twrValueBeforeCashflow sp), showDecimalT (twrPnl sp), showDecimalT (twrCashflow sp), showDecimalT (twrValueAfterCashflow sp)
+         , showDecimalT (roundTo 2 (100*rate)) ]
+       | (rate, sp) <-  subPeriods
+       ])
 
-    printf "Final unit price: %s/%s units = %s\nTotal TWR: %s%%.\nPeriod: %.2f years.\nAnnualized TWR: %.2f%%\n\n"
-      (showMixedAmount $ styleAmounts styles valueAfter) (showDecimal finalUnitBalance) (showDecimal finalUnitPrice) (showDecimal totalTWR) years annualizedTWR
+    printf "Total period TWR: %s%%.\nPeriod: %.2f years.\nAnnualized TWR: %.2f%%\n\n"
+      (showDecimal periodTWR) years annualizedTWR
 
-  return ((realToFrac totalTWR) :: Double, annualizedTWR)
+  return ((realToFrac periodTWR) :: Double, annualizedTWR)
 
 internalRateOfReturn styles showCashFlow prettyTables (OneSpan begin end valueBefore valueAfter cashFlow _pnl) = do
   let prefix = (begin, maNegate valueBefore)
@@ -301,7 +286,7 @@ internalRateOfReturn styles showCashFlow prettyTables (OneSpan begin end valueBe
       (Table
        (Tab.Group Tab.NoLine (map (Header . showDate) dates))
        (Tab.Group Tab.SingleLine [Header "Amount"])
-       (map ((:[]) . T.pack . showMixedAmount . styleAmounts styles) amts))
+       (map ((:[]) . T.pack . showMixedAmountOneLineWithoutCost False . styleAmounts styles) amts))
 
   -- 0% is always a solution, so require at least something here
   case totalCF of
@@ -324,18 +309,18 @@ interestSum referenceDay cf rate = sum $ map go cf
 
 calculateCashFlow :: WhichDate -> [Transaction] -> Query -> CashFlow
 calculateCashFlow wd trans query =
-  [ (postingDateOrDate2 wd p, pamount p) | p <- filter (matchesPosting query) (concatMap realPostings trans), maIsNonZero (pamount p) ]
+  [ (postingDateOrDate2 wd p, pamount p) | p <- concatMap (filter (matchesPosting query) . realPostings) trans, maIsNonZero (pamount p) ]
 
 total :: [Transaction] -> Query -> MixedAmount
-total trans query = sumPostings . filter (matchesPosting query) $ concatMap realPostings trans
+total trans query = sumPostings (concatMap (filter (matchesPosting query) . realPostings) trans)
 
 unMix :: MixedAmount -> Quantity
 unMix a =
-  case (unifyMixedAmount $ mixedAmountCost a) of
+  case (unifyMixedAmount a) of
     Just a' -> aquantity a'
-    Nothing -> error' $ "Amounts could not be converted to a single cost basis: " ++ show (map showAmount $ amounts a) ++
+    Nothing -> error' $ "Amounts could not be converted to a single commodity: " ++ show (map showAmount $ amounts a) ++
                "\nConsider using --value to force all costs to be in a single commodity." ++
-               "\nFor example, \"--cost --value=end,<commodity> --infer-market-prices\", where commodity is the one that was used to pay for the investment."
+               "\nFor example, \"--value=end,<commodity> --infer-market-prices\", where commodity is the one that was used for investment valuations."
 
 -- Show Decimal rounded to two decimal places, unless it has less places already. This ensures that "2" won't be shown as "2.00"
 showDecimal :: Decimal -> String
