@@ -10,6 +10,7 @@ Hledger.Utils.
 module Hledger.Cli.Utils
     (
      unsupportedOutputFormatError,
+     withJournal,
      withJournalDo,
      writeOutput,
      writeOutputLazyText,
@@ -31,12 +32,13 @@ where
 import Control.Monad.Except (ExceptT)
 import Control.Monad.IO.Class (liftIO)
 import Data.List
+import Data.List.NonEmpty qualified as NE (toList)
 import Data.Maybe
-import qualified Data.Text as T
-import qualified Data.Text.IO as T
-import qualified Data.Text.Lazy as TL
-import qualified Data.Text.Lazy.Builder as TB
-import qualified Data.Text.Lazy.IO as TL
+import Data.Text qualified as T
+import Data.Text.IO qualified as T
+import Data.Text.Lazy qualified as TL
+import Data.Text.Lazy.Builder qualified as TB
+import Data.Text.Lazy.IO qualified as TL
 import Data.Time (Day)
 import Data.Time.Clock.POSIX (POSIXTime, utcTimeToPOSIXSeconds)
 import Lens.Micro ((^.))
@@ -66,40 +68,40 @@ unsupportedOutputFormatError fmt = "Sorry, output format \""++fmt++"\" is unreco
 -- | Parse the user's specified journal file(s) as a Journal, maybe apply some
 -- transformations according to options, and run a hledger command with it.
 -- Or, throw an error.
-withJournalDo :: CliOpts -> (Journal -> IO a) -> IO a
-withJournalDo opts cmd = do
+withJournal :: CliOpts -> (Journal -> IO a) -> IO a
+withJournal opts cmd = do
   -- We kludgily read the file before parsing to grab the full text, unless
   -- it's stdin, or it doesn't exist and we are adding. We read it strictly
   -- to let the add command work.
   journalpaths <- journalFilePathFromOpts opts
-  j <- runExceptT $ journalTransform opts <$> readJournalFiles (inputopts_ opts) journalpaths
+  j <- runExceptT $ journalTransform opts <$> readJournalFiles (inputopts_ opts) (NE.toList journalpaths)
   either error' cmd j  -- PARTIAL:
 
--- | Apply some extra post-parse transformations to the journal, if enabled by options.
--- These happen after parsing and finalising the journal, but before report calculation.
--- They are, in processing order:
---
--- - pivoting account names (--pivot)
---
--- - anonymising (--anonymise).
---
+withJournalDo = withJournal
+
+-- | Apply some journal transformations, if enabled by options, that should happen late.
+-- These happen after parsing, finalising the journal, strict checks, and .latest filtering/updating,
+-- but before report calculation. They are, in processing order:
+-- --pivot, --anonymise error message, --obfuscate.
 journalTransform :: CliOpts -> Journal -> Journal
 journalTransform opts =
-      pivotByOpts opts
-  <&> anonymiseByOpts opts
+-- XXX Called by withJournal, journalReload, uiReload, withJournalCached.
+-- Could it be moved down into journalFinalise ? These steps only depend on InputOpts.
+      maybePivot opts
+  <&> maybeWarnAboutAnon opts
   <&> maybeObfuscate opts
 
--- | Apply the pivot transformation on a journal (replacing account names by a different field's value), if option is present.
-pivotByOpts :: CliOpts -> Journal -> Journal
-pivotByOpts opts =
+-- | If the --pivot option is present, replace the journal's account names by specified other values.
+maybePivot :: CliOpts -> Journal -> Journal
+maybePivot opts =
   case maybestringopt "pivot" . rawopts_ $ opts of
     Just tag -> journalPivot $ T.pack tag
     Nothing  -> id
 
 -- #2133
--- | Raise an error, announcing the rename to --obfuscate and its limitations.
-anonymiseByOpts :: CliOpts -> Journal -> Journal
-anonymiseByOpts opts =
+-- | If the --anon flag is present, raise an informative error.
+maybeWarnAboutAnon :: CliOpts -> Journal -> Journal
+maybeWarnAboutAnon opts =
   if boolopt "anon" $ rawopts_ opts
     then error' $ unlines [
        "--anon does not give privacy, and perhaps should be avoided;"
@@ -108,13 +110,18 @@ anonymiseByOpts opts =
       ]
     else id
 
--- | Apply light obfuscation to a journal, if --obfuscate is present (formerly --anon).
+-- | If the --obfuscate flag is present, apply light obfuscation to the journal data.
 maybeObfuscate :: CliOpts -> Journal -> Journal
 maybeObfuscate opts =
   if anon_ . inputopts_ $ opts
       then anon
       else id
 
+-- legacy aliases
+pivotByOpts, anonymiseByOpts :: CliOpts -> Journal -> Journal
+pivotByOpts     = maybePivot
+anonymiseByOpts = maybeObfuscate
+
 -- | Write some output to stdout or to a file selected by --output-file.
 -- If the file exists it will be overwritten.
 writeOutput :: CliOpts -> String -> IO ()
@@ -122,13 +129,14 @@ writeOutput opts s = do
   f <- outputFileFromOpts opts
   (maybe putStr writeFile f) s
 
--- | Write some output to stdout or to a file selected by --output-file.
--- If the file exists it will be overwritten. This function operates on Lazy
--- Text values.
+-- | Write some output, to a file specified by --output-file if any,
+-- otherwise to stdout.
+-- If writing to a file and the file exists, it will be overwritten.
+-- If writing to stdout, a pager is used when appropriate and possible.
 writeOutputLazyText :: CliOpts -> TL.Text -> IO ()
 writeOutputLazyText opts s = do
-  f <- outputFileFromOpts opts
-  (maybe TL.putStr TL.writeFile f) s
+  mf <- outputFileFromOpts opts
+  maybe (runPager . TL.unpack) TL.writeFile mf s
 
 -- -- | Get a journal from the given string and options, or throw an error.
 -- readJournal :: CliOpts -> String -> IO Journal
@@ -138,22 +146,21 @@ writeOutputLazyText opts s = do
 -- them has changed since last read. (If the file is standard input,
 -- this will either do nothing or give an error, not tested yet).
 -- Returns a journal or error message, and a flag indicating whether
--- it was re-read or not.  Like withJournalDo and journalReload, reads
+-- it was re-read or not.  Like withJournal and journalReload, reads
 -- the full journal, without filtering.
 journalReloadIfChanged :: CliOpts -> Day -> Journal -> ExceptT String IO (Journal, Bool)
 journalReloadIfChanged opts _d j = do
   let maybeChangedFilename f = do newer <- journalFileIsNewer j f
                                   return $ if newer then Just f else Nothing
   changedfiles <- liftIO $ catMaybes <$> mapM maybeChangedFilename (journalFilePaths j)
-  if not $ null changedfiles
-   then do
-     -- XXX not sure why we use cmdarg's verbosity here, but keep it for now
-     verbose <- liftIO isLoud
-     when (verbose || debugLevel >= 6) . liftIO $ printf "%s has changed, reloading\n" (head changedfiles)
-     newj <- journalReload opts
-     return (newj, True)
-   else
-     return (j, False)
+  case changedfiles of
+    []  -> return (j, False)
+    f:_ -> do
+      -- XXX not sure why we use cmdarg's verbosity here, but keep it for now
+      verbose <- liftIO isLoud
+      when (verbose || debugLevel >= 6) . liftIO $ printf "%s has changed, reloading\n" f
+      newj <- journalReload opts
+      return (newj, True)
 
 -- | Re-read the journal file(s) specified by options, applying any
 -- transformations specified by options. Or return an error string.
@@ -161,7 +168,7 @@ journalReloadIfChanged opts _d j = do
 journalReload :: CliOpts -> ExceptT String IO Journal
 journalReload opts = do
   journalpaths <- liftIO $ dbg6 "reloading files" <$> journalFilePathFromOpts opts
-  journalTransform opts <$> readJournalFiles (inputopts_ opts) journalpaths
+  journalTransform opts <$> readJournalFiles (inputopts_ opts) (NE.toList journalpaths)
 
 -- | Has the specified file changed since the journal was last read ?
 -- Typically this is one of the journal's journalFilePaths. These are
@@ -282,7 +289,7 @@ postingsOrTransactionsReportAsText alignAll opts itemAsText itemamt itembal repo
         startAlign = (if alignAll then id else take chunkSize) itemsWithAmounts
 
     itemsWithAmounts = map (\x -> (x, showAmt $ itemamt x, showAmt $ itembal x)) report
-    showAmt = showMixedAmountLinesB oneLine{displayColour=opts^.color__}
+    showAmt = showMixedAmountLinesB oneLineNoCostFmt{displayColour=opts^.color__}
     amt = second3
     bal = third3
 
