@@ -2,7 +2,7 @@
 
 module Main (main)
 where
-import qualified Hledger.Cli (main)
+import Hledger.Cli qualified (main)
 
 -- Have to write this explicitly for GHC 9.0.1a for some reason:
 main :: IO ()
